# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import re

import dbus

from .consts import METADATA_ALBUM, METADATA_ARTIST, METADATA_TITLE


class Metadata:
    """
    Metadata of a track

    This class helps to deal with different metadata formats defined by MPRIS1,
    MPRIS2 and OSD Lyrics. It is recommended to parse a metadata dict from D-Bus
    with `Metadata.from_dict()`.

    Metadata provides following properties: `title`, `artist`, `album`, `location`,
    `arturl`, `length`, and `tracknum`, where `length` and `tracknum` are integers,
    the others are strings.
    """

    # Possible MPRIS metadata keys, taken from
    # http://xmms2.org/wiki/MPRIS_Metadata#MPRIS_v1.0_Metadata_guidelines"""
    MPRIS1_KEYS = set(['genre', 'comment', 'rating', 'year', 'date', 'asin',
                       'puid fingerprint', 'mb track id', 'mb artist id',
                       'mb artist sort name', 'mb album id', 'mb release date',
                       'mb album artist', 'mb album artist id',
                       'mb album artist sort name', 'audio-bitrate',
                       'audio-samplerate', 'video-bitrate'])

    # Possible MPRIS2 metadata keys, taken from
    # http://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata
    MPRIS2_KEYS = set(['xesam:albumArtist', 'xesam:asText', 'xesam:audioBPM',
                       'xesam:autoRating', 'xesam:comment', 'xesam:composer',
                       'xesam:contentCreated', 'xesam:discNumber', 'xesam:firstUsed',
                       'xesam:genre', 'xesam:lastUsed', 'xesam:lyricist',
                       'xesam:useCount', 'xesam:userRating'])

    def __init__(self,
                 title=None,
                 artist=None,
                 album=None,
                 arturl=None,
                 tracknum=-1,
                 location=None,
                 length=-1,
                 extra={}):
        """
        Create a new Metadata instance.

        Arguments:
        - `title`: (string) The title of the track
        - `artist`: (string) The artist of the track
        - `album`: (string) The name of album that the track is in
        - `arturl`: (string) The URI of the picture of the cover of the album
        - `tracknum`: (int) The number of the track
        - `location`: (string) The URI of the file
        - `length`: (int) The duration of the track in milliseconds.
        - `extra`: (dict) A dict that is intend to store additional properties
                   provided by MPRIS1 or MPRIS2 DBus dicts. The MPRIS1-related
                   values will be set in the dict returned by `to_mpris1`. The
                   MPRIS2-related values are treated in a similar way.
        """
        self.title = title
        self.artist = artist
        self.album = album
        self.arturl = arturl
        self.tracknum = tracknum
        self.location = location
        self.length = length
        self._extra = extra

    def __eq__(self, other):
        """
        Two metadatas are equal if:
        - The locations are not empty and are equal, or
        - The titles, artists and albums are equal.

        See also: src/ol_metadata.c:ol_metadata_equal, thougn they aren't consistent.
        """
        if self is other:
            return True
        if self.location == other.location and self.location != '':
            return True
        for key in [METADATA_TITLE, METADATA_ARTIST, METADATA_ALBUM]:
            if getattr(self, key) != getattr(other, key):
                return False
        return True

    def to_mpris1(self):
        """
        Converts the metadata to mpris1 dict
        """
        ret = dbus.Dictionary(signature='sv')
        for k in ['title', 'artist', 'album', 'arturl', 'location']:
            if getattr(self, k) is not None:
                ret[k] = dbus.String(getattr(self, k))
        if self.tracknum >= 0:
            ret['tracknumber'] = dbus.String(self.tracknum)
        if self.length >= 0:
            ret['time'] = dbus.UInt32(self.length // 1000)
            ret['mtime'] = dbus.UInt32(self.length)
        for k, v in self._extra.items():
            if k in self.MPRIS1_KEYS and k not in ret:
                ret[k] = v
        return ret

    def to_mpris2(self):
        """
        Converts the metadata to mpris2 dict

        >>> mt = Metadata(title='Title', artist='Artist1, Artist2,Artist3',
        ...               album='Album', arturl='file:///art/url',
        ...               location='file:///path/to/file', length=123,
        ...               tracknum=456,
        ...               extra={ 'title': 'Fake Title',
        ...                       'xesam:album': 'Fake Album',
        ...                       'xesam:useCount': 780,
        ...                       'xesam:userRating': 1.0,
        ...                       'custom value': 'yoooooo',
        ...                       })
        >>> dict = mt.to_mpris2()
        >>> print(dict['xesam:title'])
        Title
        >>> print(dict['xesam:artist'])
        [dbus.String('Artist1'), dbus.String('Artist2'), dbus.String('Artist3')]
        >>> print(dict['xesam:url'])
        file:///path/to/file
        >>> print(dict['mpris:artUrl'])
        file:///art/url
        >>> print(dict['mpris:length'])
        123
        >>> print(dict['xesam:trackNumber'])
        456
        >>> print(dict['xesam:userRating'])
        1.0
        >>> 'custom value' in dict
        False
        >>> mt2 = Metadata.from_dict(dict)
        >>> print(mt2.title)
        Title
        >>> print(mt2.artist)
        Artist1, Artist2, Artist3
        >>> print(mt2.album)
        Album
        >>> print(mt2.location)
        file:///path/to/file
        """
        ret = dbus.Dictionary(signature='sv')
        mpris2map = {'title': 'xesam:title',
                     'album': 'xesam:album',
                     'arturl': 'mpris:artUrl',
                     'location': 'xesam:url',
                     }
        for k in ['title', 'album', 'arturl', 'location']:
            if getattr(self, k) is not None:
                ret[mpris2map[k]] = dbus.String(getattr(self, k))
        if self.artist is not None:
            ret['xesam:artist'] = [dbus.String(v.strip()) for v in self.artist.split(',')]
        if self.length >= 0:
            ret['mpris:length'] = dbus.Int64(self.length)
        if self.tracknum >= 0:
            ret['xesam:trackNumber'] = dbus.Int32(self.tracknum)
        for k, v in self._extra.items():
            if k in self.MPRIS2_KEYS and k not in ret:
                ret[k] = v
        return ret

    @classmethod
    def from_mpris2(cls, mpris2_dict):
        """
        Create a Metadata object from mpris2 metadata dict
        """
        string_dict = {'title': 'xesam:title',
                       'album': 'xesam:album',
                       'arturl': 'mpris:artUrl',
                       'location': 'xesam:url',
                       }
        string_list_dict = {'artist': 'xesam:artist'}
        kargs = {}
        for k, v in string_dict.items():
            if v in mpris2_dict:
                kargs[k] = mpris2_dict[v]
        for k, v in string_list_dict.items():
            if v in mpris2_dict:
                kargs[k] = ', '.join(mpris2_dict[v])
        if 'xesam:trackNumber' in mpris2_dict:
            kargs['tracknum'] = int(mpris2_dict['xesam:trackNumber'])
        if 'mpris:length' in mpris2_dict:
            kargs['length'] = int(mpris2_dict['mpris:length'])
        kargs['extra'] = mpris2_dict
        return cls(**kargs)

    @classmethod
    def from_dict(cls, dbusdict):
        """
        Create a Metadata object from a D-Bus dict object.

        The D-Bus dict object can be MPRIS1 metadata or MPRIS2 metadata format. If
        the dict both compatable with MPRIS1 and MPRIS2, MPRIS1 will be used.

        >>> title = 'Title'
        >>> artist = 'Artist'
        >>> arturl = 'file:///art/url'
        >>> location = 'file:///location'
        >>> tracknumber = 42
        >>> md1 = Metadata.from_dict({'title': title,
        ...                           'artist': artist,
        ...                           'arturl': arturl,
        ...                           'location': location,
        ...                           'tracknumber': str(tracknumber) + '/2'})
        >>> md1.title == title
        True
        >>> md1.artist == artist
        True
        >>> md1.arturl == arturl
        True
        >>> md1.location == location
        True
        >>> md1.tracknum == tracknumber
        True
        >>> md2 = Metadata.from_dict({'xesam:title': title,
        ...                           'xesam:artist': [artist],
        ...                           'mpris:artUrl': arturl,
        ...                           'xesam:url': location,
        ...                           'xesam:trackNumber': tracknumber})
        >>> md2.title == title
        True
        >>> md2.artist == artist
        True
        >>> md2.arturl == arturl
        True
        >>> md2.location == location
        True
        >>> md2.tracknum == tracknumber
        True
        >>> md3 = Metadata.from_dict({'title': title,
        ...                           'artist': artist,
        ...                           'arturl': arturl,
        ...                           'location': location,
        ...                           'tracknumber': str(tracknumber) + '/2',
        ...                           'xesam:title': title + '1',
        ...                           'xesam:artist': [artist + '1', '1'],
        ...                           'mpris:artUrl': arturl + '1',
        ...                           'xesam:url': location + '1',
        ...                           'xesam:trackNumber': tracknumber + 1})
        >>> md3.title == title
        True
        >>> md3.artist == artist
        True
        >>> md3.arturl == arturl
        True
        >>> md3.location == location
        True
        >>> md3.tracknum == tracknumber
        True
        >>> timedict = {'time': 10, 'mtime': 20, 'mpris:length': 3000}
        >>> Metadata.from_dict(timedict).length
        20
        >>> del timedict['mtime']
        >>> Metadata.from_dict(timedict).length
        3
        >>> del timedict['mpris:length']
        >>> Metadata.from_dict(timedict).length
        10000
        """
        string_dict = {'title': ['title', 'xesam:title'],
                       'album': ['album', 'xesam:album'],
                       'arturl': ['arturl', 'mpris:artUrl'],
                       'artist': ['artist'],
                       'location': ['location', 'xesam:url'],
                       }
        string_list_dict = {'artist': 'xesam:artist',
                            }
        kargs = {}
        for k, v in string_dict.items():
            for dict_key in v:
                if dict_key in dbusdict:
                    kargs[k] = dbusdict[dict_key]
                    break
        # artist
        for k, v in string_list_dict.items():
            if k not in kargs and v in dbusdict:
                kargs[k] = ', '.join(dbusdict[v])
        # tracknumber
        if 'tracknumber' in dbusdict:
            tracknumber = dbusdict['tracknumber']
            if isinstance(tracknumber, int):
                # the specification requires tracknumber be a string. However,
                # tracknumber in audacious is int32
                kargs['tracknum'] = tracknumber
            else:
                if not re.match(r'\d+(/\d+)?', tracknumber):
                    logging.warning('Malfromed tracknumber: %s', tracknumber)
                else:
                    kargs['tracknum'] = int(dbusdict['tracknumber'].split('/')[0])
        if 'tracknum' not in kargs and 'xesam:trackNumber' in dbusdict:
            kargs['tracknum'] = int(dbusdict['xesam:trackNumber'])

        # length
        if 'mtime' in dbusdict:
            kargs['length'] = dbusdict['mtime']
        elif 'mpris:length' in dbusdict:
            kargs['length'] = dbusdict['mpris:length'] // 1000
        elif 'time' in dbusdict:
            kargs['length'] = dbusdict['time'] * 1000
        kargs['extra'] = dbusdict
        return cls(**kargs)

    def __str__(self):
        attrs = ['title', 'artist', 'album', 'location', 'length']
        attr_value = ['  %s: %s' % (key, getattr(self, key)) for key in attrs]
        return 'metadata:\n' + '\n'.join(attr_value)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import dbus.exceptions


class BaseError(dbus.exceptions.DBusException):
    """ Base class for raising an exception through D-Bus
    """

    def __init__(self, *args, **kwargs):
        """

        Arguments:
        - `name`: The name of exception to send through D-Bus. If not set,
          'org.osdlyrics.Error.' + the name of the class (without 'Error'
          suffix, if any) will be returned.
        - `*args`:
        - `*kwargs`:
        """
        dbus_error_name = kwargs.pop('name', None)
        if dbus_error_name is None:
            error_name = self.__class__.__name__
            if error_name.endswith('Error'):
                error_name = error_name[:-len('Error')]
            dbus_error_name = 'org.osdlyrics.Error.' + error_name
        kwargs['name'] = dbus_error_name
        super().__init__(*args, **kwargs)


class PatternException(Exception):
    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from abc import ABCMeta
import logging
import sys
import xml.etree.ElementTree as xet

import dbus
import dbus.exceptions
import dbus.service
from gi.repository import GLib

from .property import Property

INTROSPECT_ENCODING = 'unicode'


class ObjectTypeCls(dbus.service.InterfaceType, ABCMeta):
    def __init__(cls, name, bases, dct):
        property_dict = {}
        for k, v in dct.items():
            if isinstance(v, Property):
                property_dict.setdefault(v.interface, {})[v.__name__] = v

        property_table = getattr(cls, '_dbus_property_table', {})

        for base in bases:
            if hasattr(base, '_dbus_property_table'):
                modulename = base.__module__ + '.' + base.__name__
                for interface, props in list(base._dbus_property_table[modulename].items()):
                    clsprops = property_dict.setdefault(interface, {})
                    clsprops.update(props)

        cls._dbus_property_table = property_table
        property_table[cls.__module__ + '.' + cls.__name__] = property_dict
        super(ObjectTypeCls, cls).__init__(name, bases, dct)


ObjectType = ObjectTypeCls('ObjectType', (dbus.service.Object, ), {})


class Object(ObjectType):
    """ DBus object wrapper which provides DBus property support
    """
    # __metaclass__ = ObjectType

    def __init__(self, conn=None, object_path=None, bus_name=None):
        """
        Either conn or bus_name is required; object_path is also required.

        Arguments:
        - `conn`: (dbus.connection.Connection or None) - The connection on which
           to export this object.

           If None, use the Bus associated with the given bus_name. If there is no
           bus_name either, the object is not initially available on any Connection.

           For backwards compatibility, if an instance of dbus.service.BusName is
           passed as the first parameter, this is equivalent to passing its
           associated Bus as conn, and passing the BusName itself as bus_name.

        - `object_path`: (str or None) - A D-Bus object path at which to make this
           Object available immediately. If this is not None, a conn or bus_name
           must also be provided.

        - `bus_name`: (dbus.service.BusName or None) - Represents a well-known name
           claimed by this process. A reference to the BusName object will be held
           by this Object, preventing the name from being released during this
           Object's lifetime (unless it's released manually).
        """
        dbus.service.Object.__init__(self, conn=conn,
                                     object_path=object_path,
                                     bus_name=bus_name)
        self._changed_props = {}
        self._prop_change_timer = None

    def _prop_changed_timeout_cb(self):
        self._prop_change_timer = None
        changed_props = {}
        for k, v in self._changed_props.items():
            iface = getattr(self.__class__, k).interface
            changed_props.setdefault(iface, {'changed': {}, 'invalidated': []})
            if v:
                changed_props[iface]['changed'][k] = getattr(self, k)
            else:
                changed_props[iface]['invalidated'].append(k)
        self._changed_props = {}
        for k, v in changed_props.items():
            self.PropertiesChanged(k, v['changed'], v['invalidated'])
        return False

    def _property_set(self, prop_name, emit_with_value):
        """ Callback for properties when a new value is set

        This method is called by properties of type osdlyrics.dbus.Property

        Arguments:
        - `prop_name`:
        - `changed`:
        """
        self._changed_props[prop_name] = emit_with_value
        if not self._prop_change_timer:
            self._prop_change_timer = GLib.idle_add(self._prop_changed_timeout_cb)

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                         in_signature='ss',
                         out_signature='v')
    def Get(self, iface_name, prop_name):
        """ DBus property getter

        Arguments:
        - `iface_name`:
        - `prop_name`:
        """
        prop = getattr(self.__class__, prop_name, None)
        if isinstance(prop, Property) and \
                (not iface_name or prop.interface == iface_name):
            return getattr(self, prop_name)
        raise dbus.exceptions.DBusException('No property of %s.%s' % (iface_name, prop_name))

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                         in_signature='ssv',
                         out_signature='')
    def Set(self, iface_name, prop_name, value):
        """ DBus property setter

        Arguments:
        - `iface_name`:
        - `prop_name`:
        - `value`:
        """
        prop = getattr(self.__class__, prop_name, None)
        if isinstance(prop, Property) and \
                (not iface_name or prop.interface == iface_name):
            prop.dbus_set(self, value)
        else:
            raise dbus.exceptions.DBusException('No property of %s.%s' % (iface_name, prop_name))

    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                         in_signature='s',
                         out_signature='a{sv}')
    def GetAll(self, iface_name):
        """ List all DBus properties

        Arguments:
        - `iface_name`:
        """
        ret = {}
        property_dict = self._dbus_property_table[self.__class__.__module__ + '.' + self.__class__.__name__]

        if iface_name != '' and iface_name in property_dict:
            for prop_name, prop in property_dict[iface_name].items():
                if prop.readable:
                    ret[prop_name] = prop.__get__(self)
        elif iface_name == '':
            for prop_list in property_dict.values():
                for prop_name, prop in prop_list.items():
                    if prop.readable:
                        ret[prop_name] = prop.__get__(self)
        return ret

    @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE,
                         signature='sa{sv}as')
    def PropertiesChanged(self, iface_name, changed_props, invalidated_props):
        logging.debug('%s changed: %s invalidated: %s', iface_name, changed_props, invalidated_props)
        pass

    @dbus.service.method(dbus.service.INTROSPECTABLE_IFACE, in_signature='', out_signature='s',
                         path_keyword='object_path', connection_keyword='connection')
    def Introspect(self, object_path, connection):
        """
        Patch for dbus.service.Object to add property introspection data
        """
        xml = dbus.service.Object.Introspect(self, object_path, connection)
        property_dict = self._dbus_property_table[self.__class__.__module__ + '.' + self.__class__.__name__]
        if property_dict == {}:
            return xml
        node = xet.XML(xml)
        iface_list = node.findall('interface')
        appended_iface = set()
        for iface in iface_list:
            iface_name = iface.get('name')
            if iface_name in property_dict:
                for prop_name, prop in property_dict[iface_name].items():
                    iface.append(_property2element(prop))
                appended_iface.add(iface_name)
        for iface_name, prop_list in property_dict.items():
            if iface_name in appended_iface:
                continue
            iface = xet.Element('interface', name=iface_name)
            for prop in prop_list.values():
                iface.append(_property2element(prop))
            node.append(iface)
        return '<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"\n "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">\n' + \
            xet.tostring(node, encoding=INTROSPECT_ENCODING)


def property(type_signature,
             dbus_interface,
             emit_change=True,
             readable=True,
             writeable=True):
    """
    Decorator to define dbus properties as a class member.

    To use this decorator, define your class and member as below::

       class Obj(osdlyrics.dbusext.service.Object):
            def __init__(*args, **kwargs):
                osdlyrics.dbusext.service.Object.__init__(*args, **kwargs)
                self._x = 0

            @osdlyrics.dbusext.service.property(
                type_signature='i',
                dbus_interface='org.example.Interface')
            def x(self):
                return self._x

            @x.setter
            def x(self, value):
                if self._x != value:
                    self._x = value
                    return True
                else:
                    return False

    You can use the dbus property member as a property of python

    Arguments:
    - `type_signature`: (string) The D-Bus type signature of the property
    - `dbus_interface`: (string) The D-Bus interface of the property
    - `emit_change`: (boolean or string) Whether to emit change with
                    `PropertiesChanged` D-Bus signal when the property is set.
                    Possible values are boolean value True or False, or a string
                    'invalidates'.
    - `readable`: Whether the property is able to visit with `Get` D-Bus method.
    - `writeable`: Whether the property is able to write with `Set` D-Bus method.
                   A property is writeable only when `writeable` is set to True and
                   a setter function is set.
    """
    def dec_handler(fget):
        """
        Arguments:
        - `fget`: getter function
        """
        return Property(type_signature=type_signature,
                        dbus_interface=dbus_interface,
                        emit_change=emit_change,
                        name=fget.__name__,
                        readable=readable,
                        writeable=writeable,
                        fget=fget)
    return dec_handler


def _property2element(prop):
    """
    Convert an osdlyrics.dbusext.service.Property object to
    an xml.etree.ElementTree.ElementTree object.
    """
    access = ''
    if prop.readable:
        access += 'read'
    if prop.writeable and (callable(prop._fset) or callable(prop._dbusset)):
        access += 'write'
    elem = xet.Element('property',
                       name=prop.__name__,
                       type=prop.type_signature,
                       access=access)
    if prop.emit_change != 'true':
        annotation = xet.Element('annotation',
                                 name='org.freedesktop.DBus.Property.EmitsChangedSignal',
                                 value=prop.emit_change)
        elem.append(annotation)
    return elem


def test():
    BUS_NAME = 'org.example.test'
    IFACE = 'org.example.test'
    PATH = '/org/example/test'
    DEFAULT_VALUE = 'default value of x'

    class TestObj(Object):
        def __init__(self, loop):
            Object.__init__(self, conn=dbus.SessionBus(), object_path=PATH)
            self._loop = loop
            self._foo = DEFAULT_VALUE
            self._bar = 'yes'
            self._baz = 'baz'

        @property(type_signature='s', dbus_interface=IFACE)
        def foo(self):
            return self._foo

        @foo.setter
        def foo(self, value):
            if value != self._foo:
                self._foo = value
                return True
            else:
                return False

        @property(type_signature='s',
                  dbus_interface='another.iface',
                  readable=False, writeable=True, emit_change='invalidates')
        def bar(self):
            return self._bar

        @bar.setter
        def bar(self, value):
            self._bar = value

        @property(type_signature='s',
                  dbus_interface='another.iface')
        def baz(self):
            return self._baz

        @baz.dbus_setter
        def baz(self, value):
            self._baz = value
            return True

        @baz.setter
        def baz(self, value):
            raise Exception('Shouldn\'t be called')

        @dbus.service.method(dbus_interface=IFACE,
                             in_signature='',
                             out_signature='')
        def Quit(self):
            self._loop.quit()

    def get_reply_handler(expected_value):
        def handler(value):
            if value != expected_value:
                logging.warning('Get failed, expect %s but %s got', expected_value, value)
            else:
                logging.debug('Get succesful')
        return handler

    def get_all_reply_handler(expected_dict):
        def handler(value):
            for k, v in value.items():
                if k not in expected_dict:
                    logging.warning('GetAll: unexpected key %s', k)
                elif v != expected_dict[k]:
                    logging.warning('GetAll: expected value of key %s is %s but %s got', k, expected_dict[k], v)
            for k in expected_dict:
                if k not in value:
                    logging.warning('GetAll: missing key %s', k)
            logging.debug('GetAll finished')
        return handler

    def set_reply_handler():
        logging.debug('Set succeed')

    def error_handler(e):
        logging.error('Error %s', e)

    def introspect_reply_handler(xml):
        # print xml
        pass

    def msg_handler(msg):
        def handler(*args, **kwargs):
            logging.debug(msg)
        return handler

    def test_timeout():
        proxy = conn.get_object(BUS_NAME, PATH)
        proxy.GetAll('',
                     reply_handler=get_all_reply_handler({'foo': DEFAULT_VALUE, 'baz': 'baz'}),
                     error_handler=error_handler)
        proxy.GetAll('another.iface',
                     reply_handler=get_all_reply_handler({'baz': 'baz'}),
                     error_handler=error_handler)
        proxy.Get('', 'foo',
                  reply_handler=get_reply_handler(DEFAULT_VALUE),
                  error_handler=error_handler)
        proxy.Set('', 'foo', 'new value of x',
                  reply_handler=set_reply_handler,
                  error_handler=error_handler)
        proxy.Get('', 'foo',
                  reply_handler=get_reply_handler('new value of x'),
                  error_handler=error_handler)
        proxy.Introspect(reply_handler=introspect_reply_handler,
                         error_handler=error_handler)
        proxy.Set('another.iface', 'foo', 'should not success',
                  reply_handler=msg_handler('Set error, should raise an error'),
                  error_handler=msg_handler('Set succesful, an expected exception'))
        proxy.Set('another.iface', 'baz', 'new value of baz',
                  reply_handler=set_reply_handler,
                  error_handler=error_handler)
        proxy.Get('another.iface', 'baz',
                  reply_handler=get_reply_handler('new value of baz'),
                  error_handler=error_handler)
        proxy.Set('another.iface', 'bar', 'new value of bar',
                  reply_handler=set_reply_handler,
                  error_handler=error_handler)
        proxy.Get('another.iface', 'bar',
                  reply_handler=get_reply_handler('new value of bar'),
                  error_handler=error_handler)
        return False

    class TestObjSub(TestObj):
        def __init__(self, loop):
            TestObj.__init__(self, loop)

    from gi.repository import GLib
    from dbus.mainloop.glib import DBusGMainLoop
    loop = GLib.MainLoop()
    dbus_mainloop = DBusGMainLoop()
    conn = dbus.SessionBus(mainloop=dbus_mainloop)
    bus_name = dbus.service.BusName(BUS_NAME, conn)  # noqa: F841
    testobj = TestObjSub(loop)  # noqa: F841
    GLib.timeout_add(100, test_timeout)
    loop.run()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import http.client


class InvalidRequestException(Exception):

    def __init__(self, message):
        self._message = message

    def __str__(self):
        return repr(self._message)


class HttpError(Exception):
    """
    Error response
    """

    def __init__(self, code, message=''):
        self.code = code
        self.message = message


class NotFoundError(HttpError):
    """
    HTTP Error 404
    """

    def __init__(self, message=''):
        HttpError.__init__(self, http.client.NOT_FOUND, message)


class BadRequestError(HttpError):
    """
    HTTP Error 400
    """

    def __init__(self, message=''):
        HttpError.__init__(self, http.client.BAD_REQUEST, message)


class PlayerNotFoundError(Exception):
    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager
import logging

import dbus

from osdlyrics.metadata import Metadata
from osdlyrics.player_proxy import (CAPS, REPEAT, STATUS, BasePlayer,
                                    BasePlayerProxy, PlayerInfo)

MPRIS1_INTERFACE = 'org.freedesktop.MediaPlayer'
MPRIS1_PREFIX = 'org.mpris.'

# These constants map flags/enums from MPRIS1-specific values to OSDLyrics values.
CAPS_MAP = {
    1 << 0: CAPS.NEXT,
    1 << 1: CAPS.PREV,
    1 << 2: CAPS.PAUSE,
    1 << 3: CAPS.PLAY,
    1 << 4: CAPS.SEEK,
}
STATUS_MAP = {
    0: STATUS.PLAYING,
    1: STATUS.PAUSED,
    2: STATUS.STOPPED,
}


class ProxyObject(BasePlayerProxy):
    """ The DBus object for MPRIS1 player proxy
    """

    def __init__(self):
        """
        """
        super().__init__('Mpris1')

    def _get_player_from_bus_names(self, names):
        return [
            PlayerInfo.from_name(name[len(MPRIS1_PREFIX):])
            for name in names
            if name.startswith(MPRIS1_PREFIX) and not name.startswith(MPRIS1_PREFIX + 'MediaPlayer2.')
        ]

    def do_list_active_players(self):
        return self._get_player_from_bus_names(map(str, self.connection.list_names()))

    def do_list_supported_players(self):
        return self.do_list_activatable_players()

    def do_list_activatable_players(self):
        players = self._get_player_from_bus_names(map(str, self.connection.list_activatable_names()))
        return players

    def do_connect_player(self, player_name):
        player = Mpris1Player(self, player_name)
        return player


class Mpris1Player(BasePlayer):
    def __init__(self, proxy, player_name):
        super().__init__(proxy, player_name)
        self._signals = []
        self._name_watch = None
        self._status_tuple = None, None, None, None
        self._use_cached_status = False
        try:
            self._player = dbus.Interface(self.connection.get_object(MPRIS1_PREFIX + player_name,
                                                                     '/Player'),
                                          MPRIS1_INTERFACE)
            mpris1_service_name = MPRIS1_PREFIX + player_name
            self._signals.append(self._player.connect_to_signal('TrackChange',
                                                                self._track_change_cb))
            self._signals.append(self._player.connect_to_signal('StatusChange',
                                                                self._status_change_cb))
            self._signals.append(self._player.connect_to_signal('CapsChange',
                                                                self._caps_change_cb))
            self._name_watch = self.connection.watch_name_owner(mpris1_service_name,
                                                                self._name_lost)
        except Exception as e:
            logging.error('Fail to connect to mpris1 player %s: %s', player_name, e)
            self.disconnect()

    def _name_lost(self, name):
        if name:
            return
        self.disconnect()

    def disconnect(self):
        for handler in self._signals:
            handler.remove()
        self._signals = []
        if self._name_watch:
            self._name_watch.cancel()
            self._name_watch = None
        self._player = None
        super().disconnect()

    def next(self):
        self._player.Next()

    def prev(self):
        self._player.Prev()

    def pause(self):
        self._player.Pause()

    def stop(self):
        self._player.Stop()

    def play(self):
        self._player.Play()

    def set_repeat(self, repeat):
        if repeat in (REPEAT.TRACK, REPEAT.ALL):
            self._player.Repeat(True)
        else:
            self._player.Repeat(False)

    def get_status(self):
        status_tuple = (self._status_tuple if self._use_cached_status else
                        self._player.GetStatus())
        status, shuffle, repeat, loop = status_tuple
        return STATUS_MAP.get(status, STATUS.STOPPED)

    def get_repeat(self):
        status_tuple = (self._status_tuple if self._use_cached_status else
                        self._player.GetStatus())
        status, shuffle, repeat, loop = status_tuple
        if repeat:
            return REPEAT.TRACK
        if loop:
            return REPEAT.TRACK
        return REPEAT.NONE

    def get_shuffle(self):
        status_tuple = (self._status_tuple if self._use_cached_status else
                        self._player.GetStatus())
        status, shuffle, repeat, loop = status_tuple
        return True if shuffle == 1 else False

    def get_metadata(self):
        mt = self._player.GetMetadata()
        logging.debug(repr(mt))
        return Metadata.from_dict(mt)

    def get_caps(self):
        caps = set()
        mpris1_caps = self._player.GetCaps()
        for bit, cap in CAPS_MAP.items():
            if mpris1_caps & bit:
                caps.add(cap)
        return caps

    def set_volume(self, volume):
        volume = int(volume * 100)
        if volume < 0:
            volume = 0
        if volume > 100:
            volume = 100
        self._player.VolumeSet(volume)

    def get_volume(self):
        volume = float(self._player.VolumeGet()) / 100
        if volume > 1.0:
            volume = 1.0
        if volume < 0.0:
            volume = 0.0
        return volume

    def set_position(self, time_in_mili):
        self._player.PositionSet(time_in_mili)

    def get_position(self):
        return self._player.PositionGet()

    def _track_change_cb(self, metadata):
        metadata = Metadata.from_dict(metadata)
        self.track_changed(metadata)

    def _status_change_cb(self, status_tuple):
        status, shuffle, repeat, loop = status_tuple
        old_status, old_shuffle, old_repeat, old_loop = self._status_tuple
        self._status_tuple = status_tuple
        with self._reuse_cached_status_tuple():
            if status != old_status:
                self.status_changed()
            if repeat != old_repeat or loop != old_loop:
                self.repeat_changed()
            if shuffle != old_shuffle:
                self.shuffle_changed()

    def _caps_change_cb(self, caps):
        self.caps_changed()

    @contextmanager
    def _reuse_cached_status_tuple(self):
        try:
            self._use_cached_status = True
            yield
        finally:
            self._use_cached_status = False


def run():
    mpris1 = ProxyObject()
    mpris1.run()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Tiger Soldier <tigersoldi@gmail.com>
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

# import urlparse
import gettext
import html.parser
import http.client
import re

from osdlyrics.lyricsource import BaseLyricSourcePlugin, SearchResult
from osdlyrics.utils import get_proxy_settings, http_download

_ = gettext.gettext

XIAMI_HOST = 'www.xiami.com'
XIAMI_SEARCH_URL = '/search'
XIAMI_SONG_URL = '/song/'
XIAMI_LRC_URL = '/song/playlist/id/'
XIAMI_SEARCH_PATTERN = re.compile(r'(<a [^<]*?href="http://www.xiami.com/song/(\w+)".*?>).*?(<a [^<]*?href="http://www.xiami.com/artist/.*?>).*?(<a [^<]*?href="http://www.xiami.com/album/.*?>)', re.DOTALL)
XIAMI_ID_PATTERN = re.compile(r'<a [^<]*?onclick="tag\((\d+).*?>')
XIAMI_URL_PATTERN = re.compile(r'<lyric>(.*?)</lyric>', re.DOTALL)
TITLE_ATTR_PATTERN = re.compile(r'title="(.*?)"')

gettext.bindtextdomain('osdlyrics')
gettext.textdomain('osdlyrics')


class XiamiSource(BaseLyricSourcePlugin):
    """ Lyric source from xiami.com
    """

    def __init__(self):
        super().__init__(id='xiami', name=_('Xiami'))

    def do_search(self, metadata):
        # type: (osdlyrics.metadata.Metadata) -> List[SearchResult]
        keys = []
        if metadata.title:
            keys.append(metadata.title)
        if metadata.artist:
            keys.append(metadata.artist)
        urlkey = '+'.join(keys).replace(' ', '+')
        url = XIAMI_HOST + XIAMI_SEARCH_URL
        status, content = http_download(url=url,
                                        params={'key': urlkey},
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')
        match = XIAMI_SEARCH_PATTERN.findall(content.decode('utf8'))
        result = []
        if match:
            for title_elem, id, artist_elem, album_elem in match:
                title = TITLE_ATTR_PATTERN.search(title_elem).group(1)
                artist = TITLE_ATTR_PATTERN.search(artist_elem).group(1)
                album = TITLE_ATTR_PATTERN.search(album_elem).group(1)
                url = self.get_url(id)
                if url is not None:
                    result.append(SearchResult(title=title,
                                               artist=artist,
                                               album=album,
                                               sourceid=self.id,
                                               downloadinfo=url))
        return result

    def get_songid(self, id):
        status, content = http_download(url=XIAMI_HOST + XIAMI_SONG_URL + str(id),
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            return None
        match = XIAMI_ID_PATTERN.search(content)
        if not match:
            return None
        songid = match.group(1).strip()
        return songid

    def get_url(self, id):
        songid = self.get_songid(id)
        status, content = http_download(url=XIAMI_HOST + XIAMI_LRC_URL + str(songid),
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            return None
        match = XIAMI_URL_PATTERN.search(content)
        if not match:
            return None
        url = match.group(1).strip()
        if url.lower().endswith('.lrc'):
            return url
        else:
            return None

    def do_download(self, downloadinfo):
        # type: (Any) -> bytes
        # parts = urlparse.urlparse(downloadinfo)
        status, content = http_download(downloadinfo,
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status)
        if content:
            content = html.parser.HTMLParser().unescape(content.decode('utf-8'))
        return content.encode('utf-8')


if __name__ == '__main__':
    xiami = XiamiSource()
    xiami._app.run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Tiger Soldier <tigersoldi@gmail.com>
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
import threading

import dbus

from .app import App
from .config import Config
from .consts import (LYRIC_SOURCE_PLUGIN_INTERFACE,
                     LYRIC_SOURCE_PLUGIN_OBJECT_PATH_PREFIX)
from .dbusext.service import Object as DBusObject, property as dbus_property
from .metadata import Metadata

SEARCH_SUCCEED = 0
SEARCH_CANCELLED = 1
SEARCH_FAILED = 2

DOWNLOAD_SUCCEED = 0
DOWNLOAD_CANCELLED = 1
DOWNLOAD_FAILED = 2


def onmainthread(func):
    def decfunc(self, app, *args, **kwargs):
        def timeout_cb():
            func(self, *args, **kwargs)
            return False
        app.run_on_main_thread(timeout_cb)
    return decfunc


class SearchResult:
    """ Lyrics that match the metadata to be searched.
    """

    def __init__(self, sourceid, downloadinfo, title='', artist='', album='', comment=''):
        """

        Arguments:
        - `title`: The matched lyric title.
        - `artist`: The matched lyric artist.
        - `album`: The matched lyric album.
        - `downloadinfo`: Some additional data that is needed to download the
          lyric. Normally this value is the url or ID of the lyric.
          ``downloadinfo`` MUST be composed with basic types such as numbers,
          lists, dicts or strings so that it can be converted to D-Bus compatible
          dict with `to_dict` method.
        """
        self._title = title
        self._artist = artist
        self._album = album
        self._comment = comment
        self._sourceid = sourceid
        self._downloadinfo = downloadinfo

    def to_dict(self, ):
        """ Convert the result to a dict so that it can be sent with D-Bus.
        """
        return {'title': self._title,
                'artist': self._artist,
                'album': self._album,
                'comment': self._comment,
                'sourceid': self._sourceid,
                'downloadinfo': self._downloadinfo}


class BaseTaskThread(threading.Thread):
    """ Base thread for search or download tasks.

    Plugins MUST provide a callable object as the `target` argument in the
    initializer. The target does the task and returns the results. If task
    fails, an Exception SHOULD be raised in the target.
    """

    def __init__(self, onfinish, onerror, target, args=(), kwargs={}):
        """
        Initialize the thread. The main thread should provide to callbacks
        to notify the main thread that the thread is finished or an error
        occurs. Please note that the two callbacks are run in the new thread,
        not the main thread. So the callback may notify the main thread to
        handle the results with App.run_on_main_thread.

        Arguments:

        - `onfinish`: A callable object to be invoked when `target` finish.
          The callable SHOULD receive the values returned by `do_task`.
        - `onerror`: A callable object to be invoked when `target` raise an
          exception. The callable SHOULD receive the exception raised by
          `do_task`.
        - `target`: The callable object to be invoked by the run() method.
        - `args`: The argument tuple for the target invocation. Defaults to `()`.
        - `kwargs`: A dictionary of keyword arguments for the target invocation.
          Defaults to `{}`.
        """
        super().__init__()
        self._onfinish = onfinish
        self._onerror = onerror
        self._args = args
        self._kwargs = kwargs
        self._target = target

    def run(self):
        """ Runs the task thread. Do NOT override this method. Override
        `do_task` instead.
        """
        try:
            ret = self._target(*self._args, **self._kwargs)
            self._onfinish(ret)
        except Exception as e:
            logging.exception('Got exception in thread')
            self._onerror(e)
        import sys
        sys.stdout.flush()


class BaseLyricSourcePlugin(DBusObject):
    """ Base class for implementing a lyric source plugin
    """

    def __init__(self, id, name=None, watch_daemon=True):
        """
        Create a new lyric source instance.

        Arguments:

        - `id`: The unique ID of the lyric source plugin. The full bus
          name of the plugin will be `org.osdlyrics.LyricSourcePlugin.id`
        - `name`: (optional) The name of the plugin, which should be properly
          localized. If `name` is missing, the plugin will take `id` as its
          name.
        - `watch_daemon`: Whether to watch daemon bus.
        """
        self._id = id
        self._app = App('LyricSourcePlugin.' + id,
                        watch_daemon=watch_daemon)
        super().__init__(conn=self._app.connection,
                         object_path=LYRIC_SOURCE_PLUGIN_OBJECT_PATH_PREFIX + self._id)
        self._search_count = 0
        self._download_count = 0
        self._search_tasks = {}
        self._download_tasks = {}
        self._name = name if name is not None else id
        self._config = None

    def do_search(self, metadata):
        """
        Do the real search work by plugins. All plugins MUST implement this method.

        This method runs in a seperate thread, so don't worry about block IO.

        Parameters:

        - `metadata`: The metadata of the track to search. The type of `metadata`
          is osdlyrics.metadata.Metadata

        Returns: A list of SearchResult objects
        """
        raise NotImplementedError()

    @onmainthread
    def do_searchsuccess(self, ticket, results):
        if ticket in self._search_tasks:
            del self._search_tasks[ticket]
            dbusresults = [result.to_dict() for result in results]
            self.SearchComplete(ticket, SEARCH_SUCCEED, dbusresults)

    @onmainthread
    def do_searchfailure(self, ticket, e):
        if ticket in self._search_tasks:
            del self._search_tasks[ticket]
            logging.info('Search fail, %s', e)
            self.SearchComplete(ticket, SEARCH_FAILED, [])

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         in_signature='a{sv}',
                         out_signature='i')
    def Search(self, metadata):
        ticket = self._search_count
        self._search_count = self._search_count + 1
        thread = BaseTaskThread(onfinish=lambda result: self.do_searchsuccess(self._app, ticket, result),
                                onerror=lambda e: self.do_searchfailure(self._app, ticket, e),
                                target=self.do_search,
                                kwargs={'metadata': Metadata.from_dict(metadata)})
        self._search_tasks[ticket] = thread
        thread.start()
        return ticket

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         in_signature='i',
                         out_signature='')
    def CancelSearch(self, ticket):
        if ticket in self._search_tasks:
            del self._search_tasks[ticket]
            self.SearchComplete(ticket, SEARCH_CANCELLED, [])

    def do_download(self, downloadinfo):
        """
        Do the real download work by plugins. All plugins MUST implement this
        method.

        This method runs in a seperate thread, so don't worry about block IO.

        Parameters:

        - `downloadinfo`: The additional info taken from `downloadinfo` field in
          SearchResult objects.

        Returns: A string of the lyric content
        """
        raise NotImplementedError()

    @onmainthread
    def do_downloadsuccess(self, ticket, content):
        if ticket in self._download_tasks:
            del self._download_tasks[ticket]
            self.DownloadComplete(ticket, DOWNLOAD_SUCCEED, content)

    @onmainthread
    def do_downloadfailure(self, ticket, e):
        if ticket in self._download_tasks:
            del self._download_tasks[ticket]
            self.DownloadComplete(ticket, DOWNLOAD_FAILED, str(e).encode('utf-8'))

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         in_signature='v',
                         out_signature='i')
    def Download(self, downloadinfo):
        ticket = self._download_count
        self._download_count = self._download_count + 1
        thread = BaseTaskThread(onfinish=lambda content: self.do_downloadsuccess(self._app, ticket, content),
                                onerror=lambda e: self.do_downloadfailure(self._app, ticket, e),
                                target=self.do_download,
                                kwargs={'downloadinfo': downloadinfo})
        self._download_tasks[ticket] = thread
        thread.start()
        return ticket

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         in_signature='i',
                         out_signature='')
    def CancelDownload(self, ticket):
        if ticket in self._download_tasks:
            del self._download_tasks[ticket]
            self.DownloadComplete(ticket, DOWNLOAD_CANCELLED, '')

    @dbus_property(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                   type_signature='s')
    def Name(self):
        return self._name

    @dbus.service.signal(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         signature='iiaa{sv}')
    def SearchComplete(self, ticket, status, results):
        logging.debug('search complete: ticket: %d, status: %d', ticket, status)
        pass

    @dbus.service.signal(dbus_interface=LYRIC_SOURCE_PLUGIN_INTERFACE,
                         signature='iiay')
    def DownloadComplete(self, ticket, status, result):
        logging.debug('download complete: ticket: %d, status: %d%s', ticket, status, '' if status == DOWNLOAD_SUCCEED else ', result: %s' % result)
        pass

    def run(self):
        """
        Run the plugin as a standalone application
        """
        self._app.run()

    @property
    def app(self):
        """
        Return the App object that the plugin uses.
        """
        return self._app

    @property
    def id(self):
        """
        Return the ID of the lyric source
        """
        return self._id

    @property
    def config_proxy(self):
        if self._config is None:
            self._config = Config(self._app.connection)
        return self._config


def test():
    class DummyLyricSourcePlugin(BaseLyricSourcePlugin):
        def __init__(self):
            super().__init__(id='dummy', watch_daemon=False)

        def do_search(self, metadata):
            if metadata.title:
                logging.info('title: %s', metadata.title)
                results = [SearchResult(title=metadata.title + str(i),
                                        artist=metadata.artist + str(i),
                                        album=metadata.album + str(i),
                                        sourceid=i,
                                        downloadinfo='\n'.join((metadata.title,
                                                                metadata.artist,
                                                                metadata.album)))
                           for i in range(10)]
                return results

            raise Exception('Title must not be empty')

        def do_download(self, downloadinfo):
            return b'dummy-downloaded-content'

    search_tickets = {}
    download_tickets = {}

    def search_reply(ticket, expect_status):
        if ticket in search_tickets:
            logging.warning('Error: search ticket %d exists', ticket)
        else:
            search_tickets[ticket] = expect_status

    def search_complete_cb(ticket, status, results):
        if ticket not in search_tickets:
            logging.warning('Error! search ticket not exists')
            return

        if search_tickets[ticket] != status:
            logging.warning('Error! expect search %d with status %d but %d got',
                            ticket, search_tickets[ticket], status)
            return
        logging.debug('Search #%d with status %d', ticket, status)
        if status == 0:
            downloadinfo = results[0]['downloadinfo']
            source.Download(downloadinfo,
                            reply_handler=lambda t: download_reply(t, 0),
                            error_handler=dummy_error)
        del search_tickets[ticket]
        check_quit()

    def download_reply(ticket, expect_status):
        if ticket in download_tickets:
            logging.warning('Error: download ticket %d already exists', ticket)
        else:
            download_tickets[ticket] = expect_status

    def download_complete_cb(ticket, status, content):
        if ticket not in download_tickets:
            logging.warning('Error! download ticket not exists')
            return

        if download_tickets[ticket] != status:
            logging.warning('Error! expect download status %d but %d got', download_tickets[ticket], status)
            return
        if status == 0:
            logging.debug('Download #%d success', ticket)
            logging.debug('Downloaded content: \n%s', ''.join([chr(b) for b in content]))
        else:
            logging.warning('Download #%d fail, msg: %s', ticket, ''.join([chr(b) for b in content]))
        del download_tickets[ticket]
        check_quit()

    wait_count = [4]

    def check_quit():
        wait_count[0] -= 1
        if wait_count[0] == 0:
            app.quit()

    def dummy_error(e):
        logging.warning('Error: ' + e)

    dummysource = DummyLyricSourcePlugin()
    app = dummysource.app
    conn = app.connection
    source = conn.get_object('org.osdlyrics.LyricSourcePlugin.dummy',
                             '/org/osdlyrics/LyricSourcePlugin/dummy')
    source.connect_to_signal('SearchComplete',
                             search_complete_cb)
    source.connect_to_signal('DownloadComplete',
                             download_complete_cb)
    source.Search({'title': 'dummytitle',
                   'artist': 'dummyartist',
                   'album': 'dummyalbum'},
                  reply_handler=lambda t: search_reply(t, 0),
                  error_handler=dummy_error)
    source.Search({'foo': 'bar'},
                  reply_handler=lambda t: search_reply(t, 2),
                  error_handler=dummy_error)
    source.Download(123,
                    reply_handler=lambda t: download_reply(t, 2),
                    error_handler=dummy_error)
    app.run()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

import dbus

import osdlyrics.config
from osdlyrics.consts import (LYRIC_SOURCE_PLUGIN_INTERFACE,
                              LYRIC_SOURCE_PLUGIN_OBJECT_PATH_PREFIX)

LYRIC_SOURCE_INTERFACE = 'org.osdlyrics.LyricSource'
LYRIC_SOURCE_OBJECT_PATH = '/org/osdlyrics/LyricSource'
LYRIC_SOURCE_PLUGIN_BUS_NAME_PREFIX = 'org.osdlyrics.LyricSourcePlugin.'

STATUS_SUCCESS = 0
STATUS_CANCELLED = 1
STATUS_FAILURE = 2


def validateticket(component):
    def decorator(func):
        def dec_func(self, source_id, ticket, *args, **kwargs):
            if source_id not in self._sources:
                logging.warning('%s is not in source list', source_id)
                return
            source = self._sources[source_id]
            if ticket not in source[component]:
                logging.warning('%s is not valid %s ticket of source %s', ticket, component, source_id)
                return
            func(self, source_id, ticket, *args, **kwargs)
        return dec_func
    return decorator


class LyricSource(dbus.service.Object):
    """ Implement org.osdlyrics.LyricSource interface
    """

    def __init__(self, conn):
        super().__init__(conn=conn, object_path=LYRIC_SOURCE_OBJECT_PATH)
        self._sources = {}
        self._search_tasks = {}
        self._n_search_tickets = 0
        self._download_tasks = {}
        self._n_download_tickets = 0
        self._detect_sources()
        self._config = osdlyrics.config.Config(conn)

    def _detect_sources(self):
        for bus_name in map(str, self.connection.list_names()):
            try:
                self._connect_source(bus_name, False)
            except Exception as e:
                logging.warning('Fail to connect source %s: %s', bus_name, e)
        for bus_name in map(str, self.connection.list_activatable_names()):
            try:
                self._connect_source(bus_name, True)
            except Exception as e:
                logging.warning('Fail to connect source %s: %s', bus_name, e)

    def _connect_source(self, bus_name, activate):
        if not bus_name.startswith(LYRIC_SOURCE_PLUGIN_BUS_NAME_PREFIX):
            return
        logging.info('Connecting to lyric source %s', bus_name)
        source_id = bus_name[len(LYRIC_SOURCE_PLUGIN_BUS_NAME_PREFIX):]
        if source_id in self._sources:
            return
        if activate:
            try:
                self.connection.activate_name_owner(bus_name)
            except Exception as e:
                logging.warning('Cannot activate lyric source %s: %s', bus_name, e)
                return
        path = LYRIC_SOURCE_PLUGIN_OBJECT_PATH_PREFIX + source_id
        proxy = dbus.Interface(self.connection.get_object(bus_name, path),
                               LYRIC_SOURCE_PLUGIN_INTERFACE)
        property_iface = dbus.Interface(proxy, 'org.freedesktop.DBus.Properties')
        source = {
            'proxy': proxy,
            'name': property_iface.Get(LYRIC_SOURCE_PLUGIN_INTERFACE,
                                       'Name'),
            'id': source_id,
            'search': {},
            'download': {},
        }
        self._sources[source_id] = source
        proxy.connect_to_signal('SearchComplete',
                                lambda t, s, r: self.search_complete_cb(source_id,
                                                                        t, s, r))
        proxy.connect_to_signal('DownloadComplete',
                                lambda t, s, c: self.download_complete_cb(source_id,
                                                                          t, s, c))

    @validateticket('search')
    def search_complete_cb(self, source_id, ticket, status, results):
        logging.info('Search complete from %s, ticket: %s, status: %s, result: %s',
                     source_id, ticket, status, len(results))
        source = self._sources[source_id]
        myticket = source['search'].pop(ticket)
        if myticket not in self._search_tasks:
            return
        if status == STATUS_SUCCESS:
            mytask = self._search_tasks[myticket]
            mytask['failure'] = False
        if (status == STATUS_SUCCESS and results) or \
                status == STATUS_CANCELLED:
            self.SearchComplete(myticket, status, results)
        else:  # STATUS_FAILURE
            mytask = self._search_tasks[myticket]
            # mytask['failure'] is set to True only when all sources fail to search.
            # To ensure that, we set this value to None when task is created, and
            # set it to False when a search task succeeds, thus `not task['failure']`
            # will return True. The value is set to True only when it was not False,
            # so no search tasks from other sources succeed.
            if status == STATUS_FAILURE and mytask['failure'] is not False:
                mytask['failure'] = True
            if not mytask['sources'] or mytask['sources'][0] != source_id:
                logging.warning('Error, no source exists or source id mismatch with current id')
                self.SearchComplete(myticket, STATUS_FAILURE, results)
            else:
                mytask['sources'].pop(0)
                self._do_search(myticket)

    @validateticket('download')
    def download_complete_cb(self, source_id, ticket, status, content):
        logging.info('Download complete from %s, ticket: %s, status: %s, content length: %s',
                     source_id, ticket, status, len(content))
        source = self._sources[source_id]
        myticket = source['download'].pop(ticket)
        if myticket not in self._download_tasks:
            return
        self.DownloadComplete(myticket, status, content)

    def _get_source_proxy(self, sourceid):
        return self._sources[sourceid]['proxy']

    def _get_source_search(self, sourceid, sourceticket):
        return self._sources[sourceid]['search'][sourceticket]

    def _set_source_search(self, sourceid, sourceticket, value):
        if sourceticket in self._sources[sourceid]['search']:
            raise KeyError('ticket %d exists in source search tasks', sourceticket)
        self._sources[sourceid]['search'][sourceticket] = value

    def _del_source_search(self, sourceid, sourceticket):
        del self._sources[sourceid]['search'][sourceticket]

    def _get_source_download(self, sourceid, sourceticket):
        return self._sources[sourceid]['download'][sourceticket]

    def _set_source_download(self, sourceid, sourceticket, value):
        if sourceticket in self._sources[sourceid]['download']:
            raise KeyError('ticket %d exists in source download tasks', sourceticket)
        self._sources[sourceid]['download'][sourceticket] = value

    def _del_source_download(self, sourceid, sourceticket):
        del self._sources[sourceid]['download'][sourceticket]

    def _do_search(self, ticket):
        task = self._search_tasks[ticket]
        nextsource = None
        while task['sources']:
            if task['sources'][0] in self._sources:
                nextsource = task['sources'][0]
                break
            else:
                logging.warning('Source %s not exist', task['sources'][0])
                task['sources'].pop(0)
        if nextsource is None:
            status = STATUS_SUCCESS if not task['failure'] else STATUS_FAILURE
            self.SearchComplete(ticket, status, [])
        else:
            newticket = self._get_source_proxy(nextsource).Search(task['metadata'])
            self._set_source_search(nextsource, newticket, ticket)
            task['ticket'] = newticket
            self.SearchStarted(ticket, nextsource, self._sources[nextsource]['name'])

    @dbus.service.signal(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         signature='iiaa{sv}')
    def SearchComplete(self, ticket, status, results):
        if ticket in self._search_tasks:
            del self._search_tasks[ticket]

    @dbus.service.signal(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         signature='iiay')
    def DownloadComplete(self, ticket, status, content):
        if ticket in self._download_tasks:
            del self._download_tasks[ticket]

    @dbus.service.signal(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         signature='iss')
    def SearchStarted(self, ticket, sourceid, sourcename):
        pass

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         in_signature='a{sv}as',
                         out_signature='i')
    def Search(self, metadata, sources):
        self._n_search_tickets += 1
        ticket = self._n_search_tickets
        task = {
            'metadata': metadata,
            'sources': [str(id) for id in sources],
            'ticket': None,
            'failure': None,    # See comments in search_complete_cb()
        }
        self._search_tasks[ticket] = task
        self._do_search(ticket)
        return ticket

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         in_signature='i',
                         out_signature='')
    def CancelSearch(self, ticket):
        if ticket not in self._search_tasks:
            return
        task = self._search_tasks[ticket]
        sourceticket = task['ticket']
        sourceid = task['sources'][0]
        self._get_source_proxy(sourceid).CancelSearch(sourceticket)

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         in_signature='sv',
                         out_signature='i')
    def Download(self, source_id, downloaddata):
        if source_id not in self._sources:
            return -1
        sourceticket = self._get_source_proxy(source_id).Download(downloaddata)
        if sourceticket < 0:
            return -1
        self._n_download_tickets += 1
        ticket = self._n_download_tickets
        self._download_tasks[ticket] = {
            'ticket': sourceticket,
            'source': source_id,
        }
        self._set_source_download(source_id, sourceticket, ticket)
        return ticket

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         in_signature='i',
                         out_signature='')
    def CancelDownload(self, ticket):
        if ticket not in self._download_tasks:
            return
        task = self._download_tasks[ticket]
        sourceticket = task['ticket']
        sourceid = task['source']
        self._get_source_proxy(sourceid).CancelDownload(sourceticket)

    @dbus.service.method(dbus_interface=LYRIC_SOURCE_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListSources(self):
        enabled = self._config.get_string_list('Download/download-engine')
        sources = [
            {'id': id, 'name': v['name'], 'enabled': id in enabled}
            for id, v in self._sources.items()
        ]
        order = {id: i for i, id in enumerate(enabled)}
        return sorted(sources, key=lambda it: (-it['enabled'], order.get(it['id'], 1 << 31)))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Tiger Soldier <tigersoldi@gmail.com>
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from urllib.parse import unquote

import urllib.parse
import gettext
import html.parser
import html
import http.client
import re
import os

from osdlyrics.lyricsource import BaseLyricSourcePlugin, SearchResult
from osdlyrics.utils import get_proxy_settings, http_download

_ = gettext.gettext

MEGALOBIZ_HOST = 'https://www.megalobiz.com'
MEGALOBIZ_SEARCH_PATH = '/search/all'
MEGALOBIZ_LYRIC_PATH = '/lrc/maker/'
MEGALOBIZ_DOWNLOAD_PATH = '/lrc/increment-downloads'
# Regexes to filter webpage content (search results page, and lyric page).
MEGALOBIZ_SEARCH_RESULT_PATTERN = re.compile(r'<a class=\"entity_name\".*?id=\"([0-9]+)\".*?name=\"(.*?)\".*?href=\"(.*?)\".+?</a>', re.DOTALL)
MEGALOBIZ_TITLE_ARTIST_SPLIT_PATTERN = re.compile(r'(.*?)(\sby\s|\s\-\s)(.*)', re.DOTALL)
MEGALOBIZ_LRC_PATTERN = re.compile(r'lyrics_details.*?<span.*?>(.*?)<\/span>', re.DOTALL)

gettext.bindtextdomain('osdlyrics')
gettext.textdomain('osdlyrics')


class MegalobizSource(BaseLyricSourcePlugin):
    """ Lyric source from https://www.megalobiz.com/
    """

    def __init__(self):
        super().__init__(id='megalobiz', name=_('megalobiz'))

    def do_search(self, metadata):
        # Preparing keywords for search.
        keys = []
        if metadata.artist:
            keys.append('"' + metadata.artist + '"')
        if metadata.title:
            keys.append('"' + metadata.title + '"')
        # Joining search terms.
        urlkey = '+'.join(keys).replace(' ', '+')
        # Building the URL.
        url = MEGALOBIZ_HOST + MEGALOBIZ_SEARCH_PATH
        # Request the HTTP page, storing its status and content.
        status, content = http_download(url=url,
                                        params={'qry': urlkey},
                                        proxy=get_proxy_settings(self.config_proxy))
        # Checking against HTTP response codes.
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')
        # Filter all occurences of links belonging to search results.
        matches = MEGALOBIZ_SEARCH_RESULT_PATTERN.findall(content.decode('utf8'))
        # Populate osdlyrics' search results for the user to choose from.
        result = []
        if matches:
            for match in matches:
                # The first three groups in a match are
                # lyric_id, full_name (artist + song title) and lyric path.
                # lyric_id = match[0]
                full_name = match[1]
                path = match[2]
                # Decompose optimistically the full_name into artist and song title
                name_parts = MEGALOBIZ_TITLE_ARTIST_SPLIT_PATTERN.findall(full_name)
                if isinstance(name_parts, list) and len(name_parts) > 0:
                    name_parts = len(name_parts) > 0 and name_parts[0] or exit()
                    title = name_parts[1] == ' by ' and name_parts[0] or name_parts[2]
                    artist = name_parts[1] == ' by ' and name_parts[2] or name_parts[0]
                # In case we can't split, use the full string as title.
                else:
                    title = full_name
                    artist = ''
                title = html.unescape(title)
                artist = html.unescape(artist)
                lyric_url = MEGALOBIZ_HOST + path
                if path is not None:
                    result.append(SearchResult(title=title,
                                               artist=artist,
                                               sourceid=self.id,
                                               downloadinfo=lyric_url))
        return result

    def do_download(self, downloadinfo):
        ''' Download handler for this data source.
        '''
        status, content = http_download(downloadinfo,
                                        proxy=get_proxy_settings(self.config_proxy))
        # Checking against HTTP response codes.
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status)
        # Checking the presence of a HTTP payload.
        if content:
            content = MEGALOBIZ_LRC_PATTERN.search(content.decode('utf-8')).group(1)
            # Replacing html breaks with new lines.
            content = re.sub('<br>', "\n", content)
            # Replace html entities present in the lyric content.
            content = html.unescape(content)
        return content.encode('utf-8')


if __name__ == '__main__':
    mls = MegalobizSource()
    mls._app.run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import http.server
import json
import logging
import urllib.parse

from osdlyrics.metadata import Metadata
from osdlyrics.player_proxy import CAPS, STATUS

from error import BadRequestError, HttpError, NotFoundError
from validator import (param_enum, param_int, param_set, param_str,
                       validate_params)

PARAM_STATUS = param_enum({'playing': STATUS.PLAYING,
                           'paused': STATUS.PAUSED,
                           'stopped': STATUS.STOPPED})
PARAM_CAPS = param_set({'play': CAPS.PLAY,
                        'pause': CAPS.PAUSE,
                        'next': CAPS.NEXT,
                        'prev': CAPS.PREV,
                        'seek': CAPS.SEEK})


def parse_query(query):
    """ Parse query strings in GET or POST to a dict

    The return value is a dictionary with query keys as query names and
    values as the query values. If a query name does not have a value, the
    value to the key is True. If more than one value assigned to the query name,
    any one may be assigned to the key.

    Arguments:
    - `query`: A string like 'query1=value&query2=value'
    """
    result = urllib.parse.parse_qs(query)
    ret = {}
    for k, v in result.items():
        if v:
            ret[k] = v[0]
        else:
            ret[k] = True
    return ret


class RequestHandler(http.server.BaseHTTPRequestHandler):
    """ Handles HTTP request
    """

    server_version = 'OsdLyricsHttp/1.0'

    def _send_content(self, content):
        self.send_response(200)
        self.end_headers()
        self.wfile.write(content)

    def _send_error(self, error):
        self.send_error(error.code, error.message)

    def _processquery(self, params):
        url = urllib.parse.urlparse(self.path)
        cmd = url.path[1:]
        if hasattr(self, 'do_' + cmd):
            try:
                content = getattr(self, 'do_' + cmd)(params)
                self._send_content(content)
            except HttpError as e:
                self._send_error(e)
        else:
            self._send_error(NotFoundError('Invalid request: %s' % cmd))

    def do_GET(self):
        url = urllib.parse.urlparse(self.path)
        params = parse_query(url.query)
        self._processquery(params)

    @validate_params({'name': param_str(),
                      'caps': PARAM_CAPS,
                      })
    def do_connect(self, params):
        logging.debug('caps: %s', params['caps'])
        return json.dumps({'id': self.server.player_proxy.add_player(params['name'],
                                                                     params['caps']),
                           })

    @validate_params({'id': param_str(),
                      'timestamp': param_int(),
                      })
    def do_query(self, params):
        cmds, timestamp = self.get_player(params['id']).query(params['timestamp'])
        return json.dumps({'cmds': cmds, 'timestamp': timestamp})

    @validate_params({'id': param_str(),
                      'status': PARAM_STATUS,
                      'title': param_str(),
                      'artist': param_str(optional=True),
                      'album': param_str(optional=True),
                      'arturl': param_str(optional=True),
                      'tracknum': param_int(optional=True),
                      'length': param_int()
                      })
    def do_track_changed(self, params):
        player = self.get_player(params['id'])
        status = params['status']
        del params['status']
        del params['id']
        metadata = Metadata(**params)
        player.do_update_track(metadata)
        player.do_update_status(status)
        return ''

    @validate_params({'id': param_str(),
                      'status': PARAM_STATUS,
                      })
    def do_status_changed(self, params):
        player = self.get_player(params['id'])
        player.do_update_status(params['status'])

    @validate_params({'id': param_str(),
                      'pos': param_int()})
    def do_position_changed(self, params):
        player = self.get_player(params['id'])
        player.do_update_position(params['pos'])

    @validate_params({'id': param_str()})
    def do_disconnect(self, params):
        player = self.get_player(params['id'])
        player.disconnect()

    def get_player(self, name):
        try:
            return self.server.player_proxy.get_player(name)
        except Exception:
            raise BadRequestError('Invalid player id: %s' % name)


class HttpServer(http.server.HTTPServer):
    """
    Lyrics Http server
    """

    def __init__(self, server_address, player_proxy):
        """

        Arguments:
        - `server_address`:
        """
        http.server.HTTPServer.__init__(self,
                                        server_address,
                                        RequestHandler)
        self._player_conter = 1
        self._connected_players = {}
        self._player_proxy = player_proxy

    @property
    def player_proxy(self):
        return self._player_proxy



import gettext
import http.client
import json

from osdlyrics.lyricsource import BaseLyricSourcePlugin, SearchResult
from osdlyrics.utils import get_proxy_settings, http_download

_ = gettext.gettext

NETEASE_HOST = 'music.163.com'
NETEASE_SEARCH_URL = '/api/search/get'
NETEASE_LYRIC_URL = '/api/song/lyric'

gettext.bindtextdomain('osdlyrics')
gettext.textdomain('osdlyrics')


class NeteaseTranslatedSource(BaseLyricSourcePlugin):
    """ Lyric source from music.163.com
    """

    def __init__(self, attempt_use_translation=True):
        if not attempt_use_translation:
            super().__init__(id='netease', name=_('Netease'))
        else:
            super().__init__(id='netease_tr', name=_('Netease (TR)'))
        self.attempt_use_translation = attempt_use_translation

    def do_search(self, metadata):
        # type: (osdlyrics.metadata.Metadata) -> List[SearchResult]
        keys = []
        if metadata.title:
            keys.append(metadata.title)
        if metadata.artist:
            keys.append(metadata.artist)
        url = NETEASE_HOST + NETEASE_SEARCH_URL
        urlkey = '+'.join(keys).replace(' ', '+')
        params = 's=%s&type=1' % urlkey

        status, content = http_download(url=url,
                                        method='POST',
                                        params=params.encode('utf-8'),
                                        proxy=get_proxy_settings(self.config_proxy))

        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')

        def map_func(song):
            if song['artists']:
                artist_name = song['artists'][0]['name']
            else:
                artist_name = ''
            url = NETEASE_HOST + NETEASE_LYRIC_URL + '?id=' + str(song['id']) + '&lv=-1&kv=-1&tv=-1'
            return SearchResult(title=song['name'],
                                artist=artist_name,
                                album=song['album']['name'],
                                sourceid=self.id,
                                downloadinfo=url)

        parsed = json.loads(content.decode('utf-8'))
        result = list(map(map_func, parsed['result']['songs']))

        # If there are more than 10 songs we do a second request.
        song_count = parsed['result']['songCount']
        if song_count > 10:
            params = params + '&offset=10'
            status, content = http_download(url=url,
                                            method='POST',
                                            params=params.encode('utf-8'),
                                            proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')
        parsed = json.loads(content.decode('utf-8'))
        result = result + list(map(map_func, parsed['result']['songs']))
        return result

    def do_download(self, downloadinfo):
        # type: (Any) -> bytes
        status, content = http_download(url=downloadinfo,
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status)

        parsed = json.loads(content.decode('utf-8'))
        # Avoid processing results with no lyrics.
        if 'nolyric' in parsed or 'uncollected' in parsed:
            raise ValueError('This item has no lyrics.')

        if self.attempt_use_translation:
            lyric = parsed['tlyric']['lyric']
            if not lyric:
                lyric = parsed['lrc']['lyric']
        else:
            lyric = parsed['lrc']['lyric']

        return lyric.encode('utf-8')


if __name__ == '__main__':
    netease = NeteaseTranslatedSource()
    netease._app.run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from optparse import OptionParser

import dbus
import dbus.mainloop.glib
from dbus.mainloop.glib import DBusGMainLoop
import dbus.service
from gi.repository import GLib

from .consts import DAEMON_BUS_NAME

APP_BUS_PREFIX = 'org.osdlyrics.'


dbus.mainloop.glib.threads_init()


class AlreadyRunningException(Exception):
    """ Raised when a process with given bus name exists.
    """
    pass


class App:
    """ Basic class to create a component application for OSD Lyrics.

    The application creates a mainloop, owns a DBus name, and exits when the bus
    name of OSD Lyrics disappears.

    Once the app is created, it will parse `sys.argv`. The accepted arguments are:

     - `-w`, `--watch-daemon`: If not empty, watch the daemon with given bus name.
       The default value is the bus name of OSD Lyrics

    To create an component application owning the bus name ``org.osdlyrics.MyApp``,
    just simply follow the code below:

      app = App('MyApp')
      app.run()
    """

    def __init__(self, name, watch_daemon=True, singleton=True):
        """

        Arguments:
        - `name`: The suffix of the bus name. The full bus name is
          `org.osdlyrics.` + name
        - `watch_daemon`: Whether to watch daemon bus
        - `singleton`: If True, raise AlreadyRunningException if the bus name
                       already has an owner.
        """
        self._name = name
        self._namewatch = None
        self._watch_daemon = watch_daemon
        self._loop = GLib.MainLoop()
        self._conn = dbus.SessionBus(mainloop=DBusGMainLoop())
        self._bus_names = []
        try:
            self.request_bus_name(APP_BUS_PREFIX + name,
                                  singleton)
        except dbus.NameExistsException:
            raise AlreadyRunningException(
                'Process with bus name %s is already running' % (
                    APP_BUS_PREFIX + name))
        self._parse_options()

    def _parse_options(self):
        parser = OptionParser()
        parser.add_option('-w', '--watch-daemon',
                          dest='watch_daemon',
                          action='store',
                          default=DAEMON_BUS_NAME,
                          metavar='DAEMON_BUS_NAME',
                          help=('A well-known bus name on DBus. Exit when the'
                                ' name disappears. If set to empty string,'
                                ' this player proxy will not exit.'))
        options, args = parser.parse_args()
        if self._watch_daemon:
            self._watch_daemon_bus(options.watch_daemon)

    def _watch_daemon_bus(self, name):
        if name:
            self._namewatch = self._conn.watch_name_owner(name,
                                                          self._daemon_name_changed)

    def _daemon_name_changed(self, name):
        if not name:
            self._loop.quit()
            if (self._namewatch is not None):
                self._namewatch.cancel()
                self._namewatch = None

    @property
    def connection(self):
        """The DBus connection of the app"""
        return self._conn

    @property
    def loop(self):
        return self._loop

    def run(self):
        """
        Runs the main loop

        Return True if the loop is quited by the program. False if quited by Ctrl+C
        """
        try:
            self._loop.run()
        except KeyboardInterrupt:
            return False
        return True

    def run_on_main_thread(self, target, args=(), kwargs={}):
        """Run a callable on main thread.

        This is useful for notifying a thread is finished.
        """
        def timeout_func(*user_data):
            target(*args, **kwargs)
            return GLib.SOURCE_REMOVE
        GLib.timeout_add(0, timeout_func)

    def quit(self):
        """Quits the main loop"""
        self._loop.quit()

    def request_bus_name(self, bus_name, do_not_queue=False):
        """
        Request for additional well-known name on DBus
        """
        self._bus_names.append(dbus.service.BusName(bus_name,
                                                    self.connection,
                                                    do_not_queue=do_not_queue))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

# D-Bus names, interfaces and object paths
DAEMON_INTERFACE = 'org.osdlyrics.Daemon'
DAEMON_BUS_NAME = 'org.osdlyrics.Daemon'
DAEMON_OBJECT_PATH = '/org/osdlyrics/Daemon'
MPRIS2_PREFIX = 'org.mpris.MediaPlayer2.'
DAEMON_MPRIS2_NAME = 'org.mpris.MediaPlayer2.osdlyrics'

CONFIG_BUS_NAME = 'org.osdlyrics.Config'
CONFIG_OBJECT_PATH = '/org/osdlyrics/Config'
PLAYER_PROXY_INTERFACE = 'org.osdlyrics.PlayerProxy'
PLAYER_PROXY_OBJECT_PATH_PREFIX = '/org/osdlyrics/PlayerProxy/'
MPRIS2_PLAYER_INTERFACE = 'org.mpris.MediaPlayer2.Player'
MPRIS2_OBJECT_PATH = '/org/mpris/MediaPlayer2'
LYRIC_SOURCE_PLUGIN_INTERFACE = 'org.osdlyrics.LyricSourcePlugin'
LYRIC_SOURCE_PLUGIN_OBJECT_PATH_PREFIX = '/org/osdlyrics/LyricSourcePlugin/'

# Metadata keys
METADATA_TITLE = 'title'
METADATA_ARTIST = 'artist'
METADATA_ALBUM = 'album'
METADATA_TRACKNUM = 'tracknum'
METADATA_URI = 'location'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import datetime
import logging
import time

from gi.repository import GLib

from osdlyrics.metadata import Metadata
from osdlyrics.player_proxy import (STATUS, BasePlayer, BasePlayerProxy,
                                    PlayerInfo)
import osdlyrics.timer

import server

CONNECTION_TIMEOUT = 1000


class HttpPlayerProxy(BasePlayerProxy):
    def __init__(self):
        super().__init__('Http')
        self._server = server.HttpServer(('', 7119),
                                         self)
        self._server_watch = GLib.io_add_watch(self._server.fileno(),
                                               GLib.PRIORITY_DEFAULT,
                                               GLib.IOCondition.IN,
                                               self._handle_req)
        self._players = {}
        self._connection_timer = GLib.timeout_add(CONNECTION_TIMEOUT,
                                                  self._check_connection)
        self._player_counter = 1

    def _handle_req(self, fd, event):
        logging.debug('new request %s, %s', fd, event)
        self._server.handle_request()
        return True

    def add_player(self, name, caps):
        if name in self._players:
            name = '%s%s' % (name, self._player_counter)
            self._player_counter = self._player_counter + 1
        self._players[name] = HttpPlayer(self, name, caps)
        return name

    def remove_player(self, name):
        try:
            del self._players[name]
        except KeyError:
            pass

    def get_player(self, name):
        return self._players[name]

    def do_list_active_players(self):
        return [PlayerInfo(v.name) for v in self._players.values()]

    def do_list_supported_players(self):
        return []

    def do_list_activatable_players(self):
        return []

    def do_connect_player(self, name):
        if name in self._players:
            return self._players[name]
        return None

    def _check_connection(self):
        for player in self._players.values():
            player.check_connection()
        return True


class HttpPlayer(BasePlayer):

    def __init__(self, proxy, name, caps):
        super().__init__(proxy, name)
        self._status = STATUS.STOPPED
        self._caps = caps
        self._metadata = Metadata()
        self._last_ping = datetime.datetime.now()
        self._timer = osdlyrics.timer.Timer()
        self._cmds = []

    def _ping(self):
        self._last_ping = datetime.datetime.now()

    def check_connection(self):
        now = datetime.datetime.now()
        duration = now - self._last_ping
        if duration.total_seconds() * 1000 > CONNECTION_TIMEOUT * 2:
            logging.warning('%s connection timeout', self.name)
            self.disconnect()

    def disconnect(self):
        self.proxy.remove_player(self.name)
        BasePlayer.disconnect(self)

    def do_update_track(self, metadata):
        self._ping()
        self._metadata = metadata
        self.track_changed()
        self._timer.stop()

    def do_update_status(self, status):
        self._status = status
        if status == STATUS.STOPPED:
            self._timer.stop()
        elif status == STATUS.PAUSED:
            self._timer.pause()
        else:
            self._timer.play()
        self.status_changed()

    def do_update_position(self, pos):
        self.position_changed(pos)

    def get_metadata(self):
        return self._metadata

    def get_status(self):
        return self._status

    def get_position(self):
        return self._timer.time

    def get_caps(self):
        return self._caps

    def query(self, timestamp):
        self._ping()
        cmds = []
        i = 0
        for cmd in self._cmds:
            if cmd[0] >= timestamp:
                break
            i = i + 1
        if i > 0:
            self._cmds = self._cmds[i:]
        for cmd in self._cmds:
            cmds.append(cmd[1])
        return cmds, int(time.time() * 10)

    def play(self):
        self._add_cmd('play')

    def pause(self):
        self._add_cmd('pause')

    def prev(self):
        self._add_cmd('prev')

    def next(self):
        self._add_cmd('next')

    def stop(self):
        self._add_cmd('stop')

    def set_position(self, pos):
        self._add_cmd('seek', {'pos': pos})

    def _add_cmd(self, cmd, params={}):
        self._cmds.append((int(time.time() * 10),
                           {'cmd': cmd, 'params': params}))


if __name__ == '__main__':
    proxy = HttpPlayerProxy()
    proxy.run()
